# Disable for migrations:
# pylint: disable=missing-docstring

import uuid

import django.db.models.deletion
import django_twined.models.service_revisions
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0005_question_date_help_text"),
    ]

    operations = [
        migrations.CreateModel(
            name="ServiceRevision",
            fields=[
                ("id", models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="When this service revision was created (or, strictly, added to this db)",
                    ),
                ),
                (
                    "namespace",
                    models.SlugField(
                        default=django_twined.models.service_revisions.get_default_namespace,
                        help_text="The organisation namespace, eg 'octue'",
                        max_length=80,
                    ),
                ),
                ("name", models.SlugField(help_text="The name of the service eg 'example-service'")),
                (
                    "tag",
                    models.CharField(
                        default=django_twined.models.service_revisions.get_default_tag,
                        help_text="The service revision tag that helps to identify the unique deployment",
                        max_length=80,
                    ),
                ),
                (
                    "topic_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="DEPRECATED: The UUID of this service, used for pushing questions to the service via pub/sub",
                    ),
                ),
                (
                    "project_name",
                    models.CharField(
                        default=django_twined.models.service_revisions.get_default_project_name,
                        help_text="The name of the GCP project in which the service resides",
                        max_length=80,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ServiceUsageEvent",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("data", models.JSONField(editable=False, help_text="Event payload")),
                (
                    "kind",
                    models.CharField(
                        blank=False,
                        null=False,
                        choices=[
                            ("q-asked", "Question asked"),
                            ("q-response-updated", "Question response updated"),
                            ("q-status-updated", "Question status updated"),
                        ],
                        help_text="Event kind",
                        max_length=20,
                    ),
                ),
                ("publish_time", models.DateTimeField(blank=False, help_text="Event timestamp", null=False)),
                (
                    "question",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="service_usage_events",
                        to="django_twined.question",
                    ),
                ),
                (
                    "service_revision",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="service_usage_events",
                        to="django_twined.servicerevision",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddConstraint(
            model_name="servicerevision",
            constraint=models.UniqueConstraint(fields=("namespace", "name", "tag"), name="unique_identifier"),
        ),
        migrations.AddField(
            model_name="question",
            name="service_revision",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="questions",
                to="django_twined.servicerevision",
            ),
        ),
    ]



import json

from django.conf import settings
from django.http import JsonResponse
from django_twined.models.service_revisions import ServiceRevision, service_revision_is_latest_semantic_version


SERVICE_REVISION_IS_DEFAULT_CALLBACK = getattr(
    settings,
    "TWINED_SERVICE_REVISION_IS_DEFAULT_CALLBACK",
    service_revision_is_latest_semantic_version,
)


def service_revision(request, namespace, name):
    """Get or create a service revision. If the revision tag isn't provided when getting a service revision, the default
    service revision is returned. This is the service revision with the latest semantic version revision tag by default.

    :param django.core.handlers.wsgi.WSGIRequest request:
    :param str namespace: the namespace of the service
    :param str name: the name of the service
    :return django.http.response.JsonResponse:
    """
    if request.method == "GET":
        # TODO: We can add support for version ranges here later.
        revision_tag = request.GET.get("revision_tag")

        try:
            if revision_tag:
                service_revision = ServiceRevision.objects.get(namespace=namespace, name=name, tag=revision_tag)
            else:
                service_revision = ServiceRevision.objects.get(namespace=namespace, name=name, is_default=True)

        except ServiceRevision.DoesNotExist:
            return JsonResponse({"error": "Service revision not found."}, status=404)

        return JsonResponse(
            {
                "namespace": namespace,
                "name": name,
                "revision_tag": service_revision.tag,
                "is_default": service_revision.is_default,
            },
            status=200,
        )

    if request.method == "POST":
        body = json.loads(request.body)

        if "revision_tag" not in body:
            return JsonResponse(
                {"error": "A revision tag must be included when registering a new service revision."},
                status=400,
            )

        service_revision = ServiceRevision(namespace=namespace, name=name, tag=body["revision_tag"])

        if SERVICE_REVISION_IS_DEFAULT_CALLBACK is not None:
            service_revision.is_default = SERVICE_REVISION_IS_DEFAULT_CALLBACK(service_revision)

        service_revision.save()
        return JsonResponse({}, status=201)

    return JsonResponse({"error": "Invalid request method."}, status=405)



from django.contrib import admin
from django.shortcuts import redirect
from django.urls import reverse
from django_twined.models import Question, ServiceRevision, ServiceUsageEvent

from .fieldsets import question_basic_fieldset
from .mixins import CreatableFieldsMixin


class QuestionAdmin(admin.ModelAdmin):
    """Subclass this QuestionAdmin to get started administering your question subclasses"""

    change_form_template = "django_twined/question_changeform.html"
    search_fields = ["id", "service_revision__name"]
    list_display = ("id", "asked", "answered", "service_revision")
    list_filter = (
        "asked",
        "service_revision__namespace",
        "service_revision__name",
        "service_revision__tag",
    )
    actions = ["_launch_ask_question"]
    date_hierarchy = "asked"
    readonly_fields = ("id", "asked", "answered")
    fieldsets = (
        question_basic_fieldset,
        # question_db_input_values_fieldset,
        # question_db_output_values_fieldset,
    )

    def ask_question(self, obj):
        """Override this to ask a question using an async task queue or other method. This will ask the question directly."""
        obj.ask()

    def _launch_ask_question(self, request, queryset):
        """Handler method to ask multiple question(s)"""

        ask_count = 0
        for obj in queryset.select_subclasses():
            if obj.asked is not None:
                self.message_user(request, f"{obj!r} already asked - you can't re-ask questions")
            else:
                self.ask_question(obj)
                ask_count += 1

        if ask_count == 1:
            message = "1 question was asked"
        else:
            message = f"{ask_count} questions were asked"
        self.message_user(request, message)

    _launch_ask_question.short_description = "Ask question(s)"

    def _question_is_not_asked(self, obj=None):
        """Return true if question is unasked, otherwise false"""
        if (obj is not None) and (obj.asked is not None):
            return False
        return True

    def has_change_permission(self, request, obj=None):
        """Prevent people from changing questions after they've been asked"""
        return self._question_is_not_asked(obj)

    def has_delete_permission(self, request, obj=None):
        """Prevent people from deleting questions after they've been asked"""
        return self._question_is_not_asked(obj)

    def has_duplicate_permission(self, request, obj=None):
        """Override to prevent people from duplicating questions.
        Defaults to the same as has_add_permission
        """
        return self.has_add_permission(request)

    def render_change_form(self, request, context, *args, obj=None, **kwargs):
        """Override the change form to show question ask options"""

        # If the URL was hit with a ?duplicate=True parameter, then duplicate the object and redirect to edit the new one
        duplicate = request.GET.get("duplicate", False)
        if duplicate:
            duplicate = obj.get_duplicate()
            self.message_user(request, "Duplicated question to new ID, editing the new one")

            return redirect(
                reverse(f"admin:{duplicate._meta.app_label}_{duplicate._meta.model_name}_change", args=[duplicate.id])
            )

        context.update(
            {
                "has_duplicate_permission": self.has_duplicate_permission(request, obj),
                "show_delete": obj is not None and obj.asked is None,
                "show_duplicate": obj is not None and self.has_duplicate_permission(request, obj),
                "show_save": obj is None or obj.asked is None,
                "show_save_and_add_another": False,
                "show_save_and_ask": obj is None or obj.asked is None,
                "show_save_and_continue": False,
            }
        )
        return super().render_change_form(request, context, *args, obj=obj, **kwargs)

    def response_add(self, request, obj, *args, **kwargs):
        """Save the new object then ask the question"""
        response = super().response_add(request, obj, *args, **kwargs)
        if "_save_and_ask" in request.POST:
            self.ask_question(obj.as_subclass())
            self.message_user(request, "1 question was added and asked")
        return response

    def response_change(self, request, obj):
        """Save any edits then ask the question"""
        response = super().response_change(request, obj)
        if "_save_and_ask" in request.POST:
            self.ask_question(obj.as_subclass())
            self.message_user(request, "1 question was saved and asked")
        return response


class ServiceRevisionAdmin(CreatableFieldsMixin, admin.ModelAdmin):
    """Admin panel definition for Service Revisions"""

    search_fields = ["namespace", "name", "tag"]
    list_display = ("namespace", "name", "tag", "created")
    list_filter = ("name", "tag")
    creatable_fields = ("namespace", "name", "tag", "project_name")


class ServiceUsageEventAdmin(admin.ModelAdmin):
    """Admin panel definition for Service Usage Events"""

    search_fields = ["id", "kind", "service_revision__name", "question__id"]
    list_display = ("id", "kind", "publish_time", "service_revision", "question_id")
    date_hierarchy = "publish_time"
    readonly_fields = ("id", "kind", "publish_time", "service_revision", "question")

    def has_add_permission(self, *args, **kwargs):
        """Prevent anyone from editing the event stream"""
        return False

    def has_change_permission(self, *args, **kwargs):
        """Prevent anyone from editing the event stream"""
        return False

    def has_delete_permission(self, *args, **kwargs):
        """Prevent anyone from editing the event stream"""
        return False

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related("question").prefetch_related("service_revision")

    # TODO consider the additional display options
    #   from jsoneditor.forms import JSONEditor
    #
    #   formfield_overrides = {
    #       JSONField: {"widget": JSONEditor},
    #   }
    #
    #   list_display = ("service_revision__name", "service_revision__tag", "kind", "publish_time", "question")
    #   list_filter = ("service_revision__name", "service_revision__tag", "kind", "publish_time", "question")
    #
    #   @admin.display(ordering="service_revision__name")
    #   def service_revision__name(self, obj):
    #       """Retrieve the name of the related service_revision object"""
    #       return obj.service_revision.name


admin.site.register(Question, QuestionAdmin)
admin.site.register(ServiceRevision, ServiceRevisionAdmin)
admin.site.register(ServiceUsageEvent, ServiceUsageEventAdmin)



import logging

import packaging.version
from django.conf import settings
from django.db import models, transaction
from django_gcp.events.utils import get_event_url
from octue.cloud.pub_sub.service import Service
from octue.cloud.service_id import convert_service_id_to_pub_sub_form
from octue.resources.service_backends import get_backend

from .service_usage_events import QUESTION_RESPONSE_UPDATED


logger = logging.getLogger(__name__)

QUESTION_ASK_TIMEOUT_SECONDS = 60


def service_revision_is_latest_semantic_version(service_revision):
    """Determine if a service revision is the latest semantic version based on its revision tag.

    :param django_twined.models.service_revision.ServiceRevision service_revision: the service revision to check
    :return bool: `True` if the service revision is the latest semantic version according to its revision tag
    """
    sorted_service_revisions = sorted(
        [
            *ServiceRevision.objects.filter(namespace=service_revision.namespace, name=service_revision.name),
            service_revision,
        ],
        key=lambda revision: packaging.version.parse(revision.tag),
    )

    return sorted_service_revisions[-1] == service_revision


def get_default_namespace():
    """Return the default namespace for service revisions"""
    return settings.TWINED_DEFAULT_NAMESPACE


def get_default_project_name():
    """Return the default project name for service revisions"""
    return settings.TWINED_DEFAULT_PROJECT_NAME


def get_default_tag():
    """Return the default tag if set, otherwise 'latest'"""
    return getattr(settings, "TWINED_DEFAULT_TAG", "latest")


class AbstractServiceRevision(models.Model):
    """Abstract model to register available services in the system"""

    id = models.BigAutoField(primary_key=True, editable=False)

    is_default = models.BooleanField(
        default=False,
        help_text="Set true to use this tagged revision as the default version for this name and namespace",
    )

    created = models.DateTimeField(
        auto_now_add=True, help_text="When this service revision was created (or, strictly, added to this db)"
    )

    namespace = models.SlugField(
        max_length=80,
        default=get_default_namespace,
        blank=False,
        null=False,
        help_text="The organisation namespace, eg 'octue'",
    )

    name = models.SlugField(
        help_text="The name of the service eg 'example-service'",
    )

    tag = models.CharField(
        max_length=80,
        default=get_default_tag,
        blank=False,
        null=False,
        help_text="The service revision tag that helps to identify the unique deployment",
    )

    # GCP provider-specific - will become more flexible over time.
    project_name = models.CharField(
        max_length=80,
        default=get_default_project_name,
        help_text="The name of the GCP project in which the service resides",
    )

    class Meta:
        """Metaclass for AbstractServiceRevision"""

        abstract = True
        constraints = [
            models.UniqueConstraint(
                fields=["namespace", "name", "tag"],
                name="unique_identifier",
            ),
        ]

    @property
    def sruid(self):
        """Return the Service Revision Unique Identifier

        Comprises the namespace, name, version, and tag parameters that together uniquely identify a revision (e.g. for
        the purposes of addressing).

        Using docker image labeling as inspiration, this looks like octue/example-service:0.1.2 or
        octue/example-service:my-branch or octue/example-service:0.1.2-r1 enabling both routing to specific revisions
        and things like branches for review.
        """
        return f"{self.namespace}/{self.name}:{self.tag}"

    @property
    def topic(self):
        """Return the octue GCP topic address string.

        :return str:
        """
        return "octue.services." + convert_service_id_to_pub_sub_form(self.sruid)

    def ask(
        self,
        question_id,
        input_values=None,
        input_manifest=None,
        push_url=None,
        asker_name="django-twined",
        **kwargs,
    ):
        """Ask a question of this service revision

        Asks the question and registers a push subscription, to deliver response events on the
        question topic to an endpoint.
        :param Union[str, uuid.UUID] question_id: A uuid to refer to the question by
        :param Union[dict, None] input_values: The input values of the question
        :param Union[octue.Manifest, None] input_manifest: The input manifest of the question
        :param Union[str, None] push_url: Absolute URL to the endpoint where answers and updates should be pushed (by default this is generated using django-gcp, modifying this value will require you to provide a view and full message handling at the endpoint)
        :param str asker_name: A name for this 'root service' that's asking the question.

        ```
        ServiceRevision.objects.get(
            namespace="octue",
            name="my-service",
            tag="latest"
        ).ask(
            question_id='some-uuid',
            input_values={},
            input_manifest=Manifest()
        )
        ```

        """
        if push_url is None:
            push_url = get_event_url(
                event_kind=QUESTION_RESPONSE_UPDATED,
                event_reference=question_id,
                event_parameters={
                    "srid": self.id,  # Adding this (internal) ForeignKey enables creation of ServiceUsageEvents in one DB roundtrip, not two.
                    "sruid": self.sruid,
                },
                base_url=settings.TWINED_BASE_URL,
            )

        backend = get_backend()(project_name=self.project_name)
        asker = Service(backend=backend, name=asker_name)

        subscription, _ = asker.ask(
            service_id=self.sruid,
            question_uuid=str(question_id),
            input_values=input_values,
            input_manifest=input_manifest,
            push_endpoint=push_url,
            timeout=QUESTION_ASK_TIMEOUT_SECONDS,
            **kwargs,
        )

        return subscription, push_url

    def __str__(self):
        return self.sruid

    def __repr__(self):
        return f"Service Revision ({self.sruid})"

    def save(self, *args, **kwargs):
        """Override save method to ensure that there can be only one default service revision."""
        with transaction.atomic():
            if self.is_default:
                previous_default = ServiceRevision.objects.filter(
                    namespace=self.namespace,
                    name=self.name,
                    is_default=True,
                ).first()

                if previous_default is not None:
                    previous_default.is_default = False
                    previous_default.save()

            super().save(*args, **kwargs)


class ServiceRevision(AbstractServiceRevision):
    """Concrete model to register available service revisions in the system"""


def get_default_service_revision(namespace, name):
    """Get a default service revision to use given a namespace/name combination

    Uses the is_default property of ServiceRevisions to determine which revision should be used in the absence of a specific tag

    :param str namespace: The namespace within which the default resides
    :param str name: The name for which the default revision will be returned
    :return Union[ServiceRevision | None]: The default service revision to use, or None if there isn't one

    """

    return ServiceRevision.objects.filter(name=name, namespace=namespace, is_default=True).first()



from .receivers import receive_event
from .senders import (
    delivery_acknowledgement_received,
    exception_received,
    heartbeat_received,
    log_record_received,
    monitor_message_received,
    question_asked,
    result_received,
)


__all__ = (
    "delivery_acknowledgement_received",
    "exception_received",
    "heartbeat_received",
    "log_record_received",
    "monitor_message_received",
    "question_asked",
    "receive_event",
    "result_received",
)



from django.contrib.admin.templatetags.admin_modify import InclusionAdminNode, register, submit_row


@register.tag(name="question_ask_row")
def question_ask_row_tag(parser, token):
    """Override a the question change form row tag
    https://github.com/django/django/blob/a1e4e86f923dc8387b0a9c3025bdd5d096a6ebb8/django/contrib/admin/templatetags/admin_modify.py#L115
    """
    return InclusionAdminNode(parser, token, func=submit_row, template_name="question_ask_row.html")



# Generated by Django 3.2.13 on 2022-10-06 14:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0012_servicerevision_is_default"),
    ]

    operations = [
        migrations.AlterField(
            model_name="serviceusageevent",
            name="question",
            field=models.ForeignKey(
                editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="service_usage_events",
                to="django_twined.question",
            ),
        ),
    ]



import logging

from django.dispatch import receiver
from django_gcp.events.signals import event_received
from django_gcp.events.utils import decode_pubsub_message
from django_twined.models import QUESTION_ASKED, QUESTION_RESPONSE_UPDATED, ServiceUsageEvent
from django_twined.signals.senders import (
    delivery_acknowledgement_received,
    exception_received,
    heartbeat_received,
    log_record_received,
    monitor_message_received,
    question_asked,
    result_received,
)


logger = logging.getLogger(__name__)


@receiver(event_received)
def receive_event(sender, event_kind, event_reference, event_payload, event_parameters, **kwargs):
    """Handle question updates received from pubsub
    :param event_kind (str): A kind/variety allowing you to determine the handler to use (eg "something-update"). Required.
    :param event_reference (str): A reference value provided by the client allowing events to be sorted/filtered. Required.
    :param event_payload (dict, array): The event payload to process, already decoded.
    :param event_parameters (dict): Extra parameters passed to the endpoint using URL query parameters
    :return: None
    """

    if event_kind in (QUESTION_ASKED, QUESTION_RESPONSE_UPDATED):
        decoded = decode_pubsub_message(event_payload)

        logger.debug(
            "Storing ServiceUsageEvent from event_kind %s, event_reference %s, from sender %s - PubSub message_id %s, ordering_key %s, publish_time %s, subscription %s",
            event_kind,
            event_reference,
            sender,
            decoded["message_id"],
            decoded["ordering_key"],
            decoded["publish_time"],
            decoded["subscription"],
        )

        sue = ServiceUsageEvent.objects.create(
            data=decoded["data"],
            kind=event_kind,
            publish_time=decoded["publish_time"],
            question_id=event_reference,
            service_revision_id=event_parameters["srid"],
        )

        data_type = decoded["data"].get("type", None)

        if data_type == "delivery_acknowledgement":
            delivery_acknowledgement_received.send(sender=ServiceUsageEvent, service_usage_event=sue)

        if data_type == "exception":
            exception_received.send(sender=ServiceUsageEvent, service_usage_event=sue)

        elif data_type == "heartbeat":
            heartbeat_received.send(sender=ServiceUsageEvent, service_usage_event=sue)

        elif data_type == "log_record":
            log_record_received.send(sender=ServiceUsageEvent, service_usage_event=sue)

        elif data_type == "monitor_message":
            monitor_message_received.send(sender=ServiceUsageEvent, service_usage_event=sue)

        elif data_type == "result":
            result_received.send(sender=ServiceUsageEvent, service_usage_event=sue)

        elif event_kind == QUESTION_ASKED:
            question_asked.send(sender=ServiceUsageEvent, service_usage_event=sue)

        else:
            logger.warning("Unknown event kind:type (%s) for ServiceUsageEvent %s", data_type, sue.id)



from . import fieldsets, mixins, proxy
from .admin import QuestionAdmin, ServiceRevisionAdmin, ServiceUsageEventAdmin


__all__ = (
    "QuestionAdmin",
    "ServiceUsageEventAdmin",
    "ServiceRevisionAdmin",
    "fieldsets",
    "mixins",
    "proxy",
)



# Disable for migrations:
# pylint: disable=missing-docstring

from django.db import migrations


def forward(apps, schema_editor):
    """Creates ServiceRevisions from RegisteredServices and updates relations"""

    RegisteredService = apps.get_model("django_twined", "RegisteredService")
    ServiceRevision = apps.get_model("django_twined", "ServiceRevision")
    Question = apps.get_model("django_twined", "Question")

    # Create service revisions from the registered services
    #  (use a simple iterator, as there's only ever a few).
    for registered_service in RegisteredService.objects.all():
        # Use the default namespace, tag and project_name
        service_revision = ServiceRevision.objects.create(topic_id=registered_service.id, name=registered_service.name)

        # Bulk update questions for this service
        Question.objects.filter(registered_service=registered_service).update(service_revision=service_revision)


def reverse(apps, schema_editor):
    """No-op because it's impossible to retrive unique RegisteredServices
    from sets of ServiceRevisions; you need to revert to this point (0007) then
    manually reconstruct the RegisteredServices before continuing.

    At this stage we don't anticipate anyone ever having to do this.

    """


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0006_service_revisions_and_events"),
    ]

    operations = [migrations.RunPython(forward, reverse)]



# Disable for migrations:
# pylint: disable=missing-docstring

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0004_add_timestamps"),
    ]

    operations = [
        migrations.AlterField(
            model_name="question",
            name="answered",
            field=models.DateTimeField(
                blank=True, editable=False, help_text="When the question was answered", null=True
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="asked",
            field=models.DateTimeField(blank=True, editable=False, help_text="When the question was asked", null=True),
        ),
    ]



# Disable for migrations:
# pylint: disable=missing-docstring

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0008_registered_relations_to_nullable"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="registeredservice",
            name="abstractregisteredservice_ptr",
        ),
        migrations.RemoveField(
            model_name="question",
            name="registered_service",
        ),
        migrations.DeleteModel(
            name="AbstractRegisteredService",
        ),
        migrations.DeleteModel(
            name="RegisteredService",
        ),
    ]



import logging

from django.db import transaction
from django.db.models.query import QuerySet
from django.db.utils import IntegrityError
from octue.resources import Datafile, Dataset


logger = logging.getLogger(__name__)


class StoreComparison:
    """Tiny class simply present to hold sets of IDs for store comparisons"""

    def __init__(
        self,
        datafiles_in_store,
        ids_in_store,
        ids_already_present,
        ids_missing_from_store,
        ids_missing_from_db,
        ids_imported=None,
        ids_failed_to_import=None,
        ids_synced=None,
        ids_failed_to_sync=None,
    ):
        self.datafiles_in_store = datafiles_in_store
        self.ids_in_store = ids_in_store
        self.ids_already_present = ids_already_present
        self.ids_missing_from_store = ids_missing_from_store
        self.ids_missing_from_db = ids_missing_from_db
        self.ids_imported = ids_imported
        self.ids_failed_to_import = ids_failed_to_import
        self.ids_synced = ids_synced
        self.ids_failed_to_sync = ids_failed_to_sync

    def __str__(self):

        out = ""
        for symbol in [
            "datafiles_in_store",
            "ids_in_store",
            "ids_already_present",
            "ids_missing_from_store",
            "ids_missing_from_db",
            "ids_imported",
            "ids_failed_to_import",
            "ids_synced",
            "ids_failed_to_sync",
        ]:
            out += f"{symbol}: {getattr(self,symbol)}\n"
        return out


class DatastoreQuerySetMixin:
    """A queryset mixin enabling use of django querysets as Datasets for Octue.
    Either use DatastoreQueryset directly, or use this to mix in additional queryset methods.
    """

    def delete_files(self, include_rows=True):
        """Delete the files corresponding to the rows in this query from the datalake"""
        # Get a list of filenames to delete
        files_to_delete = self.values_list(self.model.__FILE_FIELD__, flat=True)

        # Delete the datafiles from GCS
        bucket = self.model.get_storage().bucket
        for file in files_to_delete:
            bucket.blob(file).delete()

        # Run deletion of the queryset in a transaction so the database row gets restored if the file cannot be deleted
        if include_rows:
            return self.delete()

    def __init__(self, *args, store_comparison=None, **kwargs):
        self.store_comparison = store_comparison
        super().__init__(*args, **kwargs)

    def get_dataset(self, name, queryset, **kwargs):
        """Turn a django queryset into a dataset
        The queryset must be from a model that inherits from AbstractSynchronisedDatastore

        Note that this method does evaluate the full queryset with Queryset.all(), rather than yielding entries.

        :param name: The name of the dataset
        :type name: string
        :param kwargs: Optional additional keyword arguments to the octue.Dataset() constructor, e.g. {'id': '<uuid>'}
        :type kwargs: dict
        :return: The created dataset
        :rtype: octue.resources.Dataset
        """
        return Dataset(name=name, files=[*queryset.all()], **kwargs)

    def compare_store(self, cloud_paths=None):
        """Compares the contents of the store for files whose presence is not recorded in the database and imports them.

        Ignores the queryset; this method operates identically irrespective of whether called on a filtered queryset or not.

        :param list|None cloud_paths: A list of cloud_paths to import (eg ['gs://folder/file1.txt', 'gs://folder/file2.txt'].
        Import will be limited only to these paths (if they exist and those files are not already present in the store)).

        :return tuple(s): sets of ids for the resulting (imported, already_present, failed_to_import, missing_from_store) records,
        enabling further operations (such as syncing already present records from store)
        """

        if cloud_paths is not None:
            raise ValueError("cloud_paths not implemented yet")

        # Get the store to use
        storage = self.model.get_storage()

        # Get a set of ids of all existing objects, and a dict of Datafile objects keyed on their id
        ids_in_db = set(self.model.objects.values_list("id", flat=True))

        # Get a dict of Datafile objects keyed on their id, one for each object in the store
        datafiles_in_store = dict()
        for blob in storage.client.list_blobs(bucket_or_name=storage.bucket_name):
            # Get the gs path of the object
            gs_path = f"gs://{storage.bucket_name}/{blob.name}"
            logger.debug(f"Synchronizing metadata -> database for store object {gs_path}")

            # TODO This is not scalable. It will become untenable for hundreds of objects,
            # because it fetches metadata individually so generates N requests to the store.
            # Figure out whether we can list_blobs with their metadata in one request.
            datafile = Datafile(gs_path)

            # Note - datafile.id is a UUID, not the
            # same as blob.id which is a weird form of google path string
            datafiles_in_store[datafile.id] = datafile

        # Get a set of ids of all objects existing in the store
        ids_in_store = set(datafiles_in_store.keys())

        # Ids of datafiles which are both in the store and the database
        ids_already_present = ids_in_store.intersection(ids_in_db)

        # Ids of datafiles which are in the database but not the store (require deletion)
        ids_missing_from_store = ids_in_db.difference(ids_already_present)

        # Ids of datafiles which are in the store but not the database (require creation)
        ids_missing_from_db = ids_in_store.difference(ids_already_present)

        self.store_comparison = StoreComparison(
            datafiles_in_store, ids_in_store, ids_already_present, ids_missing_from_store, ids_missing_from_db
        )

        return self

    def import_missing(self):
        """Where files exist in the datastore but not the database, import those records to the DB"""

        if self.store_comparison is None:
            self = self.compare_store()

        # Prepare new records for creation in the db
        # TODO This is not scalable, we have N database inserts for N files.
        #  We could do this quicker in a bulk_create. However, failures to
        #  insert would create an all-or-nothing problem. Perhaps we can fall
        #  back to linear create on batches that fail to bulk_create?
        failed = list()
        imported = list()
        datafiles_in_store = self.store_comparison.datafiles_in_store
        for datafile_id in self.store_comparison.ids_missing_from_db:
            # Each update has to be in a transaction or the outer transaction gets left dirty after the first failure
            with transaction.atomic():
                try:
                    self.model.from_datafile(datafiles_in_store[datafile_id])
                    imported.append(datafile_id)
                except IntegrityError as e:
                    logger.error(
                        f"Could not create record from datastore object {datafiles_in_store[datafile_id].path}. Check the metadata on that file. Error was: {str(e)}"
                    )
                    failed.append(datafile_id)

        self.store_comparison.ids_imported = set(imported)
        self.store_comparison.ids_failed_to_import = set(failed)

        return self

    def delete_missing(self):
        """Where files are absent from the datastore, but there are records in the database, delete those records"""
        if self.store_comparison is None:
            self = self.compare_store()

        self.filter(id__in=self.store_comparison.ids_missing_from_store).delete()

        return self

    def sync_metadata_from_store(self):
        """Take a queryset of files and update their metadata records from the store

        This ensures that file records (metadata) are consistent with the source of truth (the datastore).

        :param bool allow_delete: If true (the default), records of files not present in the store will be
        deleted. This defaults to true in order to follow the pattern that the datastore is the source of truth.
        """

        if self.store_comparison is None:
            self = self.compare_store()

        ids_to_sync = set(str(id) for id in self.values_list("id")[0])

        # Update records with datafile meta
        # TODO This is not scalable, we have N database inserts for N files.
        #  We could do this quicker in a bulk_update. However, failures to
        #  insert would create an all-or-nothing problem. Perhaps we can fall
        #  back to linear create on batches that fail to bulk_update?
        failed = list()
        synced = list()
        datafiles_in_store = self.store_comparison.datafiles_in_store
        for id in ids_to_sync:
            # Each update has to be in a transaction or the outer transaction gets left dirty after the first failure
            with transaction.atomic():
                try:
                    self.model.from_datafile(datafiles_in_store[id], update_db_metadata=True)
                    synced.append(id)

                except IntegrityError as e:
                    logger.error(
                        f"Could not update record from datastore object {datafiles_in_store[id].path}. Check the metadata on that file. Error was: {str(e)}"
                    )
                    failed.append(id)

        self.store_comparison.ids_synced = set(synced)
        self.store_comparison.ids_failed_to_sync = set(failed)

        return self


class DatastoreQueryset(DatastoreQuerySetMixin, QuerySet):
    pass



from django.urls import path
from django.views.decorators.csrf import csrf_exempt
from django_twined.views import service_revision


urlpatterns = [
    path(r"services/<namespace>/<name>", csrf_exempt(service_revision), name="services"),
]



from django.urls import path

from .consumers import AnalysisConsumer, ServiceConsumer


websocket_urlpatterns = [
    path(r"service/", ServiceConsumer.as_asgi()),
    path(r"analyses/", AnalysisConsumer.as_asgi()),
    path(r"analyses/<uuid:analysis_id>/", AnalysisConsumer.as_asgi()),
]



# Disable for migrations:
# pylint: disable=missing-docstring

import django_twined.models.service_revisions
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0009_remove_registered_services"),
    ]

    operations = [
        migrations.AlterField(
            model_name="servicerevision",
            name="project_name",
            field=models.CharField(
                default=django_twined.models.service_revisions.get_default_project_name,
                help_text="The name of the GCP project in which the service resides",
                max_length=80,
            ),
        ),
    ]



import json
import logging
import uuid
import warnings

from django.db import models
from octue.resources import Datafile
from octue.utils.encoders import OctueJSONEncoder

from .querysets import DatastoreQueryset


logger = logging.getLogger(__name__)


class AbstractSynchronisedDatastore(models.Model):
    """Rows in the database correspond to files in an object store, with metadata synchronized for search/querying

    The most obvious question is:
        *"Why don't I just keep the metadata in the DB table and the object in the store, like normal?"*

    Some good answers:
        - because that separates the file objects and their metadata
        - because interaction with the database is then required by every service that wants to use the data, which
            - limits scalability of the system
            - requires the dba / backend team to weigh in heavily on every application that uses or creates data
            - precludes creation of data by untrusted or third party services
            - requires the dba / backend team to design and/or have deep knowledge of the intrinsic data type contained
              in the files in order to design appropriate and consistent metadata for them

    Yes, sync to a database like this is a duplication of the metadata and that requires an additional job to maintain
    consistency, which is a pain to maintain that one synchronisation system, but not an unscalable activity.

    You can think of this like you'd think about search indexes. In order to make data items searchable you'd include
    them in a search index, which is a duplicate of metadata. Here, we're duplicating metadata in order to make our data
    items queryable, relatable to other entities AND searchable (with postgres full text search, or solutions like
    ZomboDB or django-haystack to integrate elasticsearch).

    Of course you can also use your own business logic/auth and permission systems like django guardian to control
    access and visibility of your data items however you want.

    """

    __TAG_FIELDS__ = set()
    __FILE_FIELD__ = "file"

    id = models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True)

    objects = DatastoreQueryset.as_manager()

    class Meta:
        abstract = True

    @property
    def project_name(self):
        warnings.warn(
            "'project_name' property accessed on datastore object. This shouldn't be used and instead accessed directly from the storage class",
            DeprecationWarning,
        )

        return self._storage.project_id

    @property
    def datafile(self):
        """This instance as an octue datafile, constructed without accessing the store

        The metadata on this file is set from the instance, NOT from the store, so may not be fresh.

        Use `to_datafile(update_from_store=True) to ensure metadata on the datafile is fresh, although this will involve
        a call to the store so will take a long time for many datafiles.

        """
        return self.to_datafile(update_from_store=False)

    def to_datafile(self, update_from_store=True, allow_no_location=False):
        """Create an octue datafile instance corresponding to this datastore instance
        :param bool update_from_store: Fetch the latest metadata from the store. This can be an expensive operation because
        it requires an API call to the store; but is set default True to ensure that metadata reutrned is always fresh.
        Set this value to false to instead populate metadata directly from the instance, resulting in a much quicker operation.
        Generally the best approach is to set this value to false, and put in place a process to ensure your metadata is
        correctly synced from the datalake files.
        :param bool allow_no_location: Allow creation of a datafile even if the instance has no field file value set. This
        can happen on instance creation (e.g. if a data blob is being used to create an object, and therefore doesn't have a name).
        In this case, the datafile itself will be useless in the sense that data cannot be retrieved, but creating it allows
        manipulation of metadata in a way consistent with a datafile. The main use of this is to allow creation of octue metadata
        before a file is in the store, allowing use of custom storage but providing metadata on creation of the object to prevent
        repeated calls.

        :return octue.Datafile:
        """

        # Handle event of instance has unset file or file not yet in store
        if self._location is None:
            if allow_no_location:
                path = "unknown"
            else:
                raise ValueError(
                    "Instance has no file field set, cannot create Datafile. Try saving the instance first, or if you're operating on an unsaved instance see the `allow_no_location` option."
                )

        else:
            path = self.gs_path

        if update_from_store:
            df = Datafile(path)
        else:
            df = Datafile(
                path,
                ignore_stored_metadata=False,
                tags=self.get_tags_from_instance(),
                labels=self.get_labels_from_instance(),
            )

        if df.id != str(self.id):
            logger.warning(
                "Datafile at %s has bound id %s but that does not match this instance id %s. Re-sync your database to the store.",
                self.gs_path,
                df.id,
                str(self.id),
            )

        return df

    @classmethod
    def from_datafile(cls, datafile, create_if_missing=True, update_db_metadata=True):
        """Gets database record corresponding to an octue datafile object.

        The returned instance is always updated with metadata from the datafile object, however, this updated
        instance will only be saved to the database if update_db_metadata is true, so it is possible to return a dirty
        instance from this method.

        :param datafile: Datafile object representing a cloud or local file and its metadata
        :type datafile: octue.Datafile
        :param create_if_missing: Create a database entry for this file and its metadata if it's not present
        :type create_if_missing: bool
        :param update_db_metadata: If database entry for this file is found, ensure the metadata in that row is up to date with metadata from the cloud store. If false, the instance returned may be dirty (i.e. be set with values from teh store that do not correspond to the database)
        :type update_db_metadata: bool
        :return: Instance of db model
        :rtype: AbstractSynchronisedDatastore
        """

        created = False
        try:
            instance = cls.objects.get(id=datafile.id)
            instance.update_instance_from_tags(datafile.tags)
            instance.update_instance_from_labels(datafile.labels)
            if update_db_metadata:
                instance.save()

        except cls.DoesNotExist:
            instance = cls(id=datafile.id)
            instance.update_instance_from_tags(datafile.tags)
            instance.update_instance_from_labels(datafile.labels)

            # Set the file field .name attribute directly to path (https://stackoverflow.com/a/10906037/3556110)
            getattr(instance, cls.__FILE_FIELD__).name = datafile.path_in_bucket

            if create_if_missing:
                instance.save()
                created = True

        return instance, created

    def get_tags_from_instance(self):
        """Get tags (a dict of key-value pairs) from this instance

        Override this method to customise how Datafile tags are created from model instance fields (or other data)
        """
        tags = dict((name, getattr(self, name)) for name in self.__TAG_FIELDS__)

        # TODO A ***FAR*** LESS TERRIBLE SERIALIZER THAN THIS BASED ON AVRO OR SIMILAR!!!
        for k, v in tags.items():
            if v.__class__.__name__ == "Decimal":
                tags[k] = float(v)

        return tags

    def get_labels_from_instance(self):
        """Get labels as a set of strings from this instance

        Override this method to customise how Datafile labels are created from the instance fields (or other data)
        """
        return set()

    def update_instance_from_tags(self, tags):
        """Set this instance's fields using provided tags (a dict or TagDict of key-value pairs)

        Override this method to customise how metadata gets updated to the model instance (from the Datafile)
        """
        for name in self.__TAG_FIELDS__:
            setattr(self, name, tags.get(name, None))

    def update_instance_from_labels(self, labels=None):
        """Update this instance's fields using provided labels (a set of strings)

        Override this method to customise how metadata gets updated to the model instance (from the Datafile)
        """
        pass

    @property
    def _file_field(self):
        return getattr(self, self.__FILE_FIELD__)

    @property
    def _storage(self):
        return self.get_storage()

    @property
    def _storage_settings(self):
        return self.get_storage_settings()

    @classmethod
    def get_storage(cls):
        """Return the storage object for this Datastore"""
        mapped_fields = dict((field.attname, field) for field in cls._meta.concrete_fields)
        return mapped_fields[cls.__FILE_FIELD__].storage

    @classmethod
    def get_storage_settings(cls):
        """Return a dict of settings of the storage for this Datastore"""
        mapped_fields = dict((field.attname, field) for field in cls._meta.concrete_fields)
        return mapped_fields[cls.__FILE_FIELD__].store["storage_settings"]

    @property
    def _location(self):
        """The location within the bucket of the file, including name and extension, eg 'folder/subfolder/file_name.txt'
        Note that this path may not yet actually exist within the bucket, if the instance is not yet added this is what
        *will* be saved (or, at least, what will be attempted)
        """
        if getattr(self._file_field, "name", "") != "":
            return self._file_field.name
        else:
            return None

    @property
    def gs_path(self):
        """The qualified pathway of this object on GCS, or None if the blob has no name attribute"""
        loc = self._location
        if loc is not None:
            return f"gs://{self._storage.bucket_name}/{loc}"

    def _serialise_metadata(self, metadata):
        """Encode metadata as a dictionary of JSON strings."""
        if not isinstance(metadata, dict):
            raise TypeError(f"Metadata for Google Cloud storage should be a dictionary; received {metadata!r}")

        return {key: json.dumps(value, cls=OctueJSONEncoder) for key, value in metadata.items()}

    def save(self, *args, **kwargs):
        """Override the save method to add metadata to the object saved to the cloud store"""

        # Create a hypothetical (doesn't care if it's remote or local, or never existing)
        #  datafile instance simply to generate octue-compliant metadata
        datafile = self.to_datafile(update_from_store=False, allow_no_location=True)

        # Attach metadata to the file object before saving it
        self.file.file.metadata = self._serialise_metadata(datafile.metadata())
        super().save(*args, **kwargs)



import logging
import os

import dramatiq
from django.conf import settings
from django_twined.messages import ReelMessage
from octue import Runner


logger = logging.getLogger(__name__)


INPUT_STRANDS = ("input_values", "input_manifest", "credentials", "children")


@dramatiq.actor
def asker(service_name, service_version, analysis_id, input_values=None, input_manifest=None):
    """We probably want to replace all this with GCP Pub/Sub and a set of service accounts"""

    # Acknowledge the start of the run
    ask_group_name = f"analysis-{analysis_id}"
    ReelMessage(action="analysis", status="started", value=analysis_id).group_send(ask_group_name)

    try:
        # Configure the app
        service_configuration = settings.SERVICES[service_name][service_version]
        app_path = service_configuration.pop("app_path")
        twine_file = os.path.join(app_path, "twine.json")
        runner = Runner(
            app_src=app_path,
            twine=twine_file,
            **service_configuration,
        )
        logger.debug(f"Configured Runner for analysis {analysis_id}. Running...")

        # Run the app
        try:
            analysis = runner.run(input_values=input_values, input_manifest=input_manifest)
            logger.debug(f"Completed analysis {analysis_id}. Finalising...")

            results = analysis.finalise(output_dir=os.path.join("data", "output"))
            logger.debug(f"Finalised analysis {analysis_id}")

        except Exception as e:
            # Exception caused by the analysis, not the runner. Return message to the caller.
            ReelMessage(action="ask", status="error", value=analysis_id, hints=e.args[0]).group_send(ask_group_name)
            return

        # Create the completion message
        ReelMessage(action="ask", status="complete", value=analysis_id, **results).group_send(ask_group_name)

    except Exception as e:
        # Exception caused by the analysis, not the runner. Return message to the caller.
        ReelMessage(
            action="ask",
            status="error",
            value=analysis_id,
            hints="Internal error occurred. Could not complete analysis. Your error has been logged for examination.",
        ).group_send(ask_group_name)
        raise e


def ask(analysis_id, message):
    """Start the ask process. Returns a ReelMessage that can be sent to confirm the ask has been queued."""

    # Get only the twine arguments
    kwargs = dict((k, getattr(message, k, None)) for k in INPUT_STRANDS)
    asker.send(analysis_id, **kwargs)

    return ReelMessage(action=message.action, reference=message.reference, status="queued", value=analysis_id)



import logging
import os

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer
from django.conf import settings
from django_twined.messages import ReelMessage
from django_twined.tasks import ask


logger = logging.getLogger(__name__)


class ServiceConsumer(WebsocketConsumer):
    """A high-level consumer allowing subscription to service messages, resource messages and the current twine.

    Resource messages are sent only to a subset of clients (like members of a chat room).

        TODO ACTUALLY GET THIS OUT PROPERLY - Currently uses the default

    ~~ When in danger or in doubt, run in circles, scream and shout ~~

    """

    @property
    def service_name(self):
        """The service name for this connection"""
        # NB We define this as a property so it's an instance attribute, rather than a class attribute, without
        # having to overload __init__.
        return str(self.scope["url_route"]["kwargs"].get("service_name", None))

    @property
    def service_version(self):
        """The service version for this connection

        TODO ACTUALLY GET THIS OUT PROPERLY - Currently uses the default. Figure out a good way of listing the available
         services and versions. Possibly using a service table. Also need to consider the channel layer name which
         should include the version

        """
        # NB We define this as a property so it's an instance attribute, rather than a class attribute, without
        # having to overload __init__.
        return settings.SERVICES[self.service_name]["default_version"]

    @property
    def resource_name(self):
        """The resource group name for this connection, if any"""
        # NB We define this as a property so it's an instance attribute, rather than a class attribute, without
        # having to overload __init__.
        return str(self.scope["url_route"]["kwargs"].get("resource_name", None))

    def connect(self):
        """Accept connection to the service, and send twine to the client as acknowledgement"""

        # The service name should be given in the URL somewhere, if not, don't accept the connection
        if self.service_name is None:
            self.close()

        # Add this channel to the service group (unused at present)
        # async_to_sync(self.channel_layer.group_add)(self.service_name, self.channel_name)

        # Optionally add this channel to a particular resource group, if one is found in the URL kwargs
        if self.resource_name is not None:
            async_to_sync(self.channel_layer.group_add)(self.resource_name, self.channel_name)

        # Accept the connection
        self.accept()

        # Send the twine for this service to the connected client
        #   TODO this might be better done via a conventional view.
        #   TODO Check read-configuration permissions and make the configuration available
        self._send_twine()

    def disconnect(self, code):
        """Accept a disconnection from the service and associated groups gracefully"""

        # Disconnect from the service
        async_to_sync(self.channel_layer.group_discard)(self.service_name, self.channel_name)

        # Disconnect from the resource group, if a member
        if self.resource_name is not None:
            async_to_sync(self.channel_layer.group_discard)(self.resource_name, self.channel_name)

    def receive(self, text_data=None, bytes_data=None):
        """Receive service instructions from WebSocket"""

        # Parse the incoming message data
        try:
            message = ReelMessage(src=text_data)
        except Exception:
            ReelMessage(status="error", reference=message.reference, hints="Unable to understand your message").send(
                self
            )
            return

        if message.action == "ping":
            ReelMessage(
                action="pong",
                status="success",
                reference=message.reference,
                hints="U wanna play? I can pong your ping all day.",
            ).send(self)

        if message.action == "ask":

            # Forward the inbound message to the resource group
            if self.resource_name:
                message.group_send(self.resource_name)

            ask(self.service_name, self.service_version, message, self.resource_name or self.channel_name)

            # # Forward the inbount message to the resource group
            # if self.resource_group_name:
            #     ReelMessage(src=text_data, action="asked").group_send(self.resource_group_name)
            #
            #     # TODO more elegant send API for ReelMessage so I can just provide an iterable of group names
            #     #  or *Consumer instances
            #     message.group_send(self.resource_group_name)
            # else:
            #     message.send(self)
            #
            #     # TODO more elegant send API for ReelMessage so I can just provide an iterable of group names
            #     #  or *Consumer instances
            #     message.group_send(self.resource_group_name)
            # else:
            #     message.send(self)
            #

        else:
            ReelMessage(
                action=message.action,
                status="error",
                hint="Unknown action. Perhaps you meant to send the message to the analysis consumer?",
            ).send(self)

    def _send_twine(self):
        """Refresh the client with everything it needs to know about the twine and the analyses it has going on"""

        app_path = settings.TWINED_SERVICES[self.service_name][self.service_version]["path"]
        with open(os.path.join(app_path, "twine.json")) as fp:
            twine = fp.read()

        ReelMessage(action="connect", twine=twine).send(self)



import logging

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand


logger = logging.getLogger(__name__)


VERBOSITY_MAP = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


class Command(BaseCommand):
    """Use `python manage.py help sync_data_stores` to display help for this command line administration tool"""

    help = (
        "Syncs files and their metadata from stores (defined in SETTINGS) into queryable records in the database."
        "Use in development for repopulating an empty database with the contents of the store(s),"
        "or use in production if a migration is applied to metadata stored in the database and you need to"
        "propagate that to the store"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--source-keys",
            nargs="+",
            required=False,
            dest="source_keys",
            default=None,
            help="Specify the data store keys. If none, all keys specified in settings.TWINED_DATA_STORES will be used.",
        )
        parser.add_argument(
            "--db-to-store",
            required=False,
            dest="source_keys",
            default=None,
            help="By default, metadata is synced from store to db. To do the other way arou",
        )

    def handle(self, *args, source_keys=None, **options):

        # Ensure we respect the --verbosity command option
        verbosity = int(options["verbosity"])
        logger.setLevel(VERBOSITY_MAP[verbosity])

        # Get the data sources to synchronise (by default, all defined)
        source_keys = source_keys or settings.TWINED_DATA_STORES.keys()
        stores = dict((key, settings.TWINED_DATA_STORES[key]) for key in source_keys)

        # Loop through the data sources
        for key, store in stores.items():
            logger.info("Synchronizing store -> database for %s", key)

            # Get model class and the storage instance used for its datafiles
            Model = apps.get_model(*store["model"].split("."))

            # Daisychain sync commands so the store comparison only gets run once
            Model.objects.import_missing().delete_missing().sync_metadata_from_store()



# Disable for migrations:
# pylint: disable=missing-docstring

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0007_registered_to_revision"),
    ]

    operations = [
        migrations.AlterField(
            model_name="question",
            name="registered_service",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="questions",
                to="django_twined.registeredservice",
            ),
        )
    ]



question_basic_fieldset = (
    None,
    {
        "fields": (
            "id",
            "service_revision",
            "asked",
            "answered",
        )
    },
)

question_db_input_values_fieldset = ("Input Values", {"classes": ("collapse",), "fields": ("input_values",)})
question_db_output_values_fieldset = ("Output Values", {"classes": ("collapse",), "fields": ("output_values",)})



# Generated by Django 3.2.13 on 2022-06-06 13:14

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0010_alter_servicerevision_project_name"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="servicerevision",
            name="topic_id",
        ),
    ]



# Generated by Django 3.2.13 on 2022-10-04 15:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0011_remove_servicerevision_topic_id"),
    ]

    operations = [
        migrations.AddField(
            model_name="servicerevision",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="Set true to use this tagged revision as the default version for this name and namespace",
            ),
        ),
    ]



import logging

from channels.generic.websocket import WebsocketConsumer
from django_twined.messages import ReelMessage


logger = logging.getLogger(__name__)


class AnalysisLogConsumer(WebsocketConsumer):
    """A consumer allowing logs for an analysis to be streamed in plain text directly from a python logger, then
    converted into ReelMessages which get logged to the analysis group.
    """

    @property
    def analysis_id(self):
        return str(self.scope["url_route"]["kwargs"].get("analysis_id"))

    @property
    def analysis_group_name(self):
        return f"analysis-{self.analysis_id}"

    def connect(self):
        """Accept connection to to enable log forwarding to the analysis group"""
        self.accept()

    def receive(self, text_data=None, bytes_data=None):
        """Receive log text, wrap it properly as part of the messaging system and forward it to the analysis group"""
        ReelMessage(action="log", value=text_data).group_send(self.analysis_group_name)



class CreatableFieldsMixin:
    """A mixin allowing specification of admin fields that are editable on create but not on change."""

    creatable_fields = tuple()

    # def get_fields(self, request, obj=None):
    #     """Override readonly_fields to add creatable_fields in the case of object update, but not in the case of object creation"""
    #     defaults = super().get_fields(request, obj=obj)
    #     return tuple(defaults) + self.creatable_fields

    def get_readonly_fields(self, request, obj=None):
        """Override readonly_fields to add creatable_fields in the case of object update, but not in the case of object creation"""
        defaults = super().get_readonly_fields(request, obj=obj)
        if obj:  # if we are updating an object
            defaults = tuple(defaults) + self.creatable_fields  # make sure defaults is a tuple
        return defaults



from django.dispatch import Signal


delivery_acknowledgement_received = Signal()
exception_received = Signal()
heartbeat_received = Signal()
log_record_received = Signal()
monitor_message_received = Signal()
question_asked = Signal()
result_received = Signal()



import logging
import uuid
from datetime import datetime, timezone

from django.db import models
from model_utils.managers import InheritanceManager

from ..fields import ManifestField, ValuesField


logger = logging.getLogger(__name__)


class AbstractQuestion(models.Model):
    """Abstract Base Class for a Question model to store questions asked to octue services"""

    duplicate_fields = ()

    id = models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True)
    asked = models.DateTimeField(null=True, blank=True, editable=False, help_text="When the question was asked")
    answered = models.DateTimeField(null=True, blank=True, editable=False, help_text="When the question was answered")

    class Meta:
        """Metaclass for AbstractQuestion"""

        abstract = True

    objects = InheritanceManager()

    def __str__(self):
        return f"{self.id}"

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.id})"

    def get_duplicate(self, save=True):
        """Duplicate the question instance and optionally save to the database"""
        kwargs = {}
        for field in self.duplicate_fields:
            kwargs[field] = getattr(self, field)
        # Always override the asked and answered attributes of the duplicate
        overrides = {"asked": None, "answered": None}
        duplicate = self.__class__(**kwargs, **overrides)
        if save:
            duplicate.save()
        return duplicate

    def get_input_values(self):
        """Get the input values from wherever they're stored (or compute them)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_input_values method for this class (or provide input directly to ask() method)"
        )

    def get_input_manifest(self):
        """Get the input manifest from wherever it's stored (or compute it)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_input_manifest method for this class (or provide input directly to ask() method)"
        )

    def get_service_revision(self):
        """Get the service revision from wherever it's stored (or otherwise determine it)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_service_revision method for this class (or provide input directly to ask() method)"
        )

    def as_subclass(self):
        """Returns a base question object as its inherited subclass"""
        return self.__class__.objects.get_subclass(id=self.id)

    def ask(self, save=True):
        """Ask a question to a service_revision.

        :param boolean save: If true, this question will be saved in order to update the 'asked' time.
        """

        service_revision = self.get_service_revision()
        subscription, push_url = service_revision.ask(
            question_id=self.id,
            input_values=self.get_input_values(),
            input_manifest=self.get_input_manifest(),
        )

        logger.info("Service Revision %s was asked question %s", service_revision.sruid, self.id)

        self.asked = datetime.now().replace(tzinfo=timezone.utc)

        if save:
            self.save()

        return subscription, push_url, service_revision


class Question(AbstractQuestion):
    """Stores questions asked to octue services

    This concrete model is here to link questions to the service revisions
    used to ask them. It is intended to be subclassed for specific question types,
    hence why it doesn't override the getters for input and output values.

    """

    service_revision = models.ForeignKey(
        "django_twined.ServiceRevision",
        blank=True,
        null=True,
        related_name="questions",
        on_delete=models.PROTECT,
    )

    def get_input_values(self):
        """Get the input values from wherever they're stored (or compute them)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_input_values method for this class (or provide input directly to ask() method)"
        )

    def get_input_manifest(self):
        """Get the input manifest from wherever it's stored (or compute it)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_input_manifest method for this class (or provide input directly to ask() method)"
        )

    def get_service_revision(self):
        return self.service_revision


class QuestionValuesDatabaseStorageMixin:
    """DEPRECATED - DO NOT USE
    Use this mixin to store question-and-answer values data in the actual database

    This will be deprecated as we move to using an event store for question asks and updates,
    to avoid duplicates.

    Instead, override the get_input_values() and get_output_values() methods to access
    the event store.
    """

    input_values = ValuesField(help_text="Contents of the input_values strand")

    output_values = ValuesField(help_text="Contents of the output_values strand")

    def get_input_values(self):
        """Override the get_input_values abstract class method to return the values from the database store"""
        return self.input_values

    def get_output_values(self):
        """Override the get_output_values abstract class method to return the values from the database store"""
        return self.output_values


class QuestionManifestsDatabaseStorageMixin:
    """DEPRECATED - DO NOT USE
    Use this mixin to store question-and-answer manifest data in the actual database.

    This will be deprecated as we move to event based processing, to avoid duplicates. Do not change!
    Instead, use ManyToManyFields to <Subclass>DataStore in your <Subclass>Question, and override
    the get_input_manifest() and get_output_manifest() methods.
    """

    input_manifest = ManifestField(help_text="Contents of the input_manifest strand")

    output_manifest = ManifestField(help_text="Contents of the output_manifest strand")

    def get_input_manifest(self):
        """Override the get_input_manifest abstract class method to return the manifest from the database store"""
        return self.input_manifest

    def get_output_manifest(self):
        """Override the get_output_manifest abstract class method to return the manifest from the database store"""
        return self.output_manifest



from django.contrib import admin


def register_proxy_modeladmin(modeladmin, model, name=None):
    """Registers a modeladmin entry using a proxy model class

    This allows multiple modeladmins to be registered to the same
    model class, meaning you can have different modeladmins for the
    same things.

    Based on https://stackoverflow.com/a/2228821/3556110
    """

    class Meta:
        """Define a metaclass used to construct a proxy model"""

        proxy = True
        app_label = model._meta.app_label

    attrs = {"__module__": "", "Meta": Meta}

    newmodel = type(name, (model,), attrs)

    admin.site.register(newmodel, modeladmin)

    return modeladmin



from .datastores import AbstractSynchronisedDatastore
from .questions import AbstractQuestion, Question, QuestionValuesDatabaseStorageMixin
from .service_revisions import AbstractServiceRevision, ServiceRevision, get_default_service_revision
from .service_usage_events import QUESTION_ASKED, QUESTION_RESPONSE_UPDATED, QUESTION_STATUS_UPDATED, ServiceUsageEvent


__all__ = (
    "AbstractServiceRevision",
    "AbstractSynchronisedDatastore",
    "AbstractQuestion",
    "get_default_service_revision",
    "Question",
    "QuestionValuesDatabaseStorageMixin",
    "QUESTION_ASKED",
    "QUESTION_RESPONSE_UPDATED",
    "QUESTION_STATUS_UPDATED",
    "ServiceRevision",
    "ServiceUsageEvent",
)



import logging

from django.db import models


logger = logging.getLogger(__name__)


QUESTION_ASKED = "q-asked"
QUESTION_RESPONSE_UPDATED = "q-response-updated"
QUESTION_STATUS_UPDATED = "q-status-updated"

SERVICE_USAGE_EVENT_KINDS = {
    QUESTION_ASKED: "Question asked",
    QUESTION_RESPONSE_UPDATED: "Question response updated",
    QUESTION_STATUS_UPDATED: "Question status updated",
}

SERVICE_USAGE_EVENT_KINDS_CHOICES = tuple((k, v) for k, v in SERVICE_USAGE_EVENT_KINDS.items())


class AbstractEvent(models.Model):
    """Abstract base model for storing events

    This allows creation of multiple different tables for storage of different events.

    """

    id = models.BigAutoField(primary_key=True)

    data = models.JSONField(blank=False, null=False, editable=False, help_text="Event payload")

    kind = models.CharField(
        max_length=20, null=False, blank=False, help_text="Event kind", choices=SERVICE_USAGE_EVENT_KINDS_CHOICES
    )

    publish_time = models.DateTimeField(null=False, blank=False, help_text="Event timestamp")

    class Meta:
        """AbstractEvent meta class data"""

        abstract = True


class ServiceUsageEvent(AbstractEvent):
    """Table for storing Service Usage Events

    This table will store events related to usage of octue services (for example, question asks, answers, updates and responses)
    in your main database.

    For high volume processing, you may wish not to store events at all, or you may wish to store in a different
    database such as BigQuery. More helper tools for this coming soon.

    This table relates to concrete Question and Service Revision models. Null relations are
    not allowed, meaning that the table cannot store events not generated with django-twined
    (i.e. where Service Revision and Question models do not exist). This may change in future,
    to enable absorption of questions triggered by other services.
    """

    question = models.ForeignKey(
        "django_twined.Question",
        null=False,
        blank=False,
        related_name="service_usage_events",
        on_delete=models.CASCADE,
        editable=False,
    )

    service_revision = models.ForeignKey(
        "django_twined.ServiceRevision",
        null=False,
        blank=False,
        related_name="service_usage_events",
        on_delete=models.PROTECT,
        editable=False,
    )

    def __str__(self):
        return str(self.kind)

    def __repr__(self):
        return f"Service Usage Event {self.kind}"
